#include <glib.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

typedef struct _Vert Vert;
struct _Vert {
	char *name;
	GList *edges;
};

static int size = 0;

/*make zero polynomial*/
static int *
make_poly(void)
{
	int i;
	int *p = g_malloc((size+1)*sizeof(int));
	for(i=0;i<size+1;i++)
		p[i]=0;
	return p;
}

/*sub p2 from p1*/
static void
sub_poly(int *p1,int *p2)
{
	int i;
	for(i=0;i<size+1;i++)
		p1[i]-=p2[i];
}

static void
print_vert(gpointer key, gpointer value, gpointer user_data)
{
	Vert *vert = value;
	GList *li;
	char *sep = "";
	printf("%s:",vert->name);
	for(li=vert->edges;li!=NULL;li=g_list_next(li)) {
		Vert *v = li->data;
		printf("%s%s",sep,v->name);
		sep = " ";
	}
	puts("");
}

static void
print_graph(GHashTable *graph)
{
	g_hash_table_foreach(graph,print_vert,NULL);
}

static void
copy_vert(gpointer key, gpointer value, gpointer user_data)
{
	GHashTable *new = user_data;
	Vert *ov = value;
	Vert *nv = g_new0(Vert,1);
	GList *li;
	nv->name = g_strdup(ov->name);

	for(li=ov->edges;li!=NULL;li=g_list_next(li)) {
		Vert *v = li->data;
		nv->edges = g_list_append(nv->edges,g_strdup(v->name));
	}
	g_hash_table_insert(new,nv->name,nv);
}

static void
connect_graph(gpointer key, gpointer value, gpointer user_data)
{
	Vert *vert = value;
	GList *li;
	GList *nl=NULL;
	GHashTable *hash = user_data;

	for(li=vert->edges;li!=NULL;li=g_list_next(li)) {
		Vert *v = g_hash_table_lookup(hash,li->data);
		if(v != vert && v)
			nl = g_list_append(nl,v);
		g_free(li->data);
	}
	g_list_free(vert->edges);
	vert->edges = nl;
}

static GHashTable *
copy_graph(GHashTable *old)
{
	GHashTable *new = g_hash_table_new(g_str_hash,g_str_equal);
	g_hash_table_foreach(old,copy_vert,new);
	g_hash_table_foreach(new,connect_graph,new);
	return new;
}

static void
find_first_with_edge(gpointer key, gpointer value, gpointer user_data)
{
	Vert *vert = value;
	Vert **first = user_data;
	if(*first) return;
	if(vert->edges)
		*first = vert;
}

/*remove the first edge you find from a graph*/
static void
kill_first_edge(GHashTable *gr)
{
	Vert *first = NULL;
	Vert *to;
	g_hash_table_foreach(gr,find_first_with_edge,&first);
	if(!first) return;
	
	to = first->edges->data;
	to->edges = g_list_remove(to->edges,first);
	first->edges = g_list_remove(first->edges,to);
}

static void
replace_references(gpointer key, gpointer value, gpointer user_data)
{
	Vert *vert = value;
	Vert **pass = user_data;
	GList *li;
	if(vert==pass[1] || g_list_find(vert->edges,pass[1]))
		vert->edges = g_list_remove(vert->edges,pass[0]);
	else if((li=g_list_find(vert->edges,pass[0])))
		li->data = pass[1];
}

/*remove the first edge you find from a graph and merge the two verticies*/
static void
merge_first_edge(GHashTable *gr)
{
	Vert *first = NULL;
	Vert *to;
	Vert *pass[2];
	GList *li;
	g_hash_table_foreach(gr,find_first_with_edge,&first);
	if(!first) return;
	
	to = first->edges->data;

	pass[0] = to;
	pass[1] = first;
	g_hash_table_foreach(gr,replace_references,pass);
	
	for(li=to->edges; li!=NULL; li=g_list_next(li)) {
		if(li->data!=first &&
		   !g_list_find(first->edges,li->data))
			first->edges = g_list_append(first->edges,li->data);
	}
	g_hash_table_remove(gr,to->name);
	g_free(to->name);
	g_list_free(to->edges);
	g_free(to);
}

static void
correct_graph(gpointer key, gpointer value, gpointer user_data)
{
	Vert *vert = value;
	GList *li;
	
	for(li=vert->edges;li!=NULL;li=g_list_next(li)) {
		Vert *v = li->data;
		if(!g_list_find(v->edges,vert))
			v->edges = g_list_append(v->edges,vert);
	}
}

static GHashTable *
read_graph_in(FILE *fp)
{
	char line[4096];
	GHashTable *graph;

	graph = g_hash_table_new(g_str_hash,g_str_equal);
	
	while(fgets(line,4096,fp)) {
		char *p;
		Vert *vert;
		p=strchr(line,'\n');
		if(p) *p = '\0';
		p = strtok(line," \t,;:");
		if(!p) continue;
		vert = g_new0(Vert,1);
		vert->name = g_strdup(p);
		while((p = strtok(NULL," \t,;:")))
			vert->edges = g_list_append(vert->edges, g_strdup(p));
		g_hash_table_insert(graph,vert->name,vert);
	}
	g_hash_table_foreach(graph,connect_graph,graph);
	g_hash_table_foreach(graph,correct_graph,NULL);
	return graph;
}

static void
free_vert(gpointer key,gpointer value,gpointer data)
{
	Vert *vert = value;
	g_free(vert->name);
	g_list_free(vert->edges);
	g_free(vert);
}

static void
free_graph(GHashTable *graph)
{
	g_hash_table_foreach(graph,free_vert,NULL);
	g_hash_table_destroy(graph);
}

static int *
calculate_chromatic_poly(GHashTable *graph)
{
	int *poly,*poly2;
	GHashTable *gp;	
	GHashTable *gpp;	

	Vert *first = NULL;
	
	/*see if there even is an edge*/
	g_hash_table_foreach(graph,find_first_with_edge,&first);
	if(!first) {
		poly = make_poly();
		poly[g_hash_table_size(graph)]++;
		return poly;
	}

	gp = copy_graph(graph);
	kill_first_edge(gp);

	gpp = copy_graph(graph);
	merge_first_edge(gpp);

	poly = calculate_chromatic_poly(gp);
	free_graph(gp);

	poly2 = calculate_chromatic_poly(gpp);
	free_graph(gpp);

	sub_poly(poly,poly2);
	g_free(poly2);

	return poly;
}

static long
int_power(int x,int e)
{
	long r = 1;
	while(e>0) {
		if(e & 0x1) { /*if odd*/
			r *= x;
			e--;
		} else { /* even */
			x *= x;
			e >>= 1; /*e /= 2*/
		}
	}
	return r;
}

static long
calculate_poly(int *poly,int x)
{
	int i;
	long r = 0;
	for(i=0;i<size+1;i++) {
		if(poly[i] == 0) continue;
		r += poly[i]*int_power(x,i);
	}
	return r;
}

static int
calculate_chromatic_num(int *poly)
{
	int i;
	for(i=1;i<INT_MAX;i++) {
		if(calculate_poly(poly,i)!=0)
			return i;
	}
	printf("Could not find a chromatic number between 1 and %d\n",INT_MAX);
	return -1;
}

int
main(void)
{
	int i,any;
	int *poly;
	GHashTable *graph = NULL;

	graph = read_graph_in(stdin);

	puts("The Graph:");
	print_graph(graph);
	
	size = g_hash_table_size(graph);
	
	/*calculate polynomial*/
	poly = calculate_chromatic_poly(graph);

	/*print out polynomial*/
	puts("Chromatic polynomial:");
	any = FALSE;
	for(i=size;i>=0;i--) {
		if(poly[i]==0) continue;
		if(any && poly[i]>0)
			printf(" + ");
		else if(any && poly[i]<0)
			printf(" - ");
		else if(poly[i]<0)
			printf("-");
		if(poly[i]!=1 && poly[i]!=-1)
			printf("%d",poly[i]<0?-poly[i]:poly[i]);
		if(i==1)
			printf("x");
		else if(i>1)
			printf("x^%d",i);
		any = TRUE;
	}
	if(!any)
		puts("0");
	else
		puts("");
	
	printf("Chromatic number:\n%d\n",calculate_chromatic_num(poly));

	return 0;
}
